import logging

import oauth2

import requests
import factual.common.session

class Session(factual.common.session.BaseSession):
	_url_pat = "%(api_base)s/t/%(table_id)s/%(action)s?%(query)s"
	def __init__(
				self, 
				consumer_key=None, 
				consumer_secret=None, 
				api_key=None, # deprecated
				api_base="http://api.v3.factual.com"
				):
		'''
		Construct a Session object to hold OAuth credentials. Provide an OAuth consumer 
		key and consumer secret to use with the Factual API.

		n.b. Providing only an api_key or only a consumer_key will fall back to 
		     authenticating requests via the KEY query string parameter, skipping 
			 OAuth entirely. This behavior is deprecated in the v3 API. 
		'''
		self.consumer_key = consumer_key if consumer_key else api_key
		self.consumer_secret = consumer_secret
		self.api_base = api_base

		if api_key:
			logging.warn("The api_key parameter is deprecated in the v3 API. Use OAuth.")

		if self.consumer_secret:
			self.consumer = oauth2.Consumer(consumer_key, consumer_secret)
			self.oauth_signature_method = oauth2.SignatureMethod_HMAC_SHA1()
		else:
			self.consumer = None
			logging.warn("Non-OAuth requests are deprecated in the v3 API. Pass both a consumer key and secret to use OAuth.")

	def get_url(self, request):
		query = request.get_query()
		if not self.consumer:
			# fall back to api_key behavior
			query = "%s&KEY=%s" % (query, self.consumer_key,)
		return self._url_pat % {
				'api_base': self.api_base,
				'table_id': request.table,
				'action':   request.action,
				'query':    query
			}
	def get_headers(self, request, defaults={}):
		headers = {}
		if self.consumer:
			# only do OAuth if we have a consumer_secret to use, else fall back to api_key in URL

			# todo: check that Factual API is always GET/blank body/not form encoded
			oauth_request = oauth2.Request.from_consumer_and_token(
								consumer=self.consumer,
								token=None,

								http_method="GET",
								body="",
								is_form_encoded=False,

								http_url=self.get_url(request),
								parameters=None
								)

			# pass None as the Token since Factual's API is 2-legged OAuth only
			oauth_request.sign_request(self.oauth_signature_method, self.consumer, None)
			headers.update(oauth_request.to_header())
		headers.update(defaults)
		return headers
	def read(self, table):
		'''Build a Read request on the provided Table type using this Session. See the requests module for details.'''
		return requests.Read(self, table)




'''
Helper functions to handle Factual point of interest category queries and mappings.

These functions assume that the user (a) wants to limit results to the smallest
set of catgories matching a predefined list and (b) wants to map Factual's category
labels into his own category labelling system. 
'''

import shared_filter_helpers as ops

def make_category_filter(categories, blank=True):
	'''
	Generates a dict representing a Factual filter matching any of the categories
	passed. 

	The resulting filter uses $bw "begins with" operators to return all matching
	subcategories. Because of this, passing a top level category removes the need
	to pass any of its subcategories. 

	Conversely, specifying subcategories will not restrict results as expected
	if a prefix of those subcategories is also provided. For example:
		make_category_filter(["Food & Beverage", "Food & Beverage > Cheese"])
	is the same as 
		make_category_filter(["Food & Beverage"])
	
	To minimize the size of the filters sent to Factual, make_category_filters 
	identifies redundant subcategories and removes them. 

	Note that because of this prefix matching, queries may return rows from unwanted
	subcategories. It may be necessary for you to filter out these records after 
	the Factual request. 

	Specify blank=True to include items without a category set.
	'''
	categories = [category.strip() for category in categories]

	# find shortest prefixes
	categories.sort()
	redundant_categories = set()
	prefix_candidate = None
	for category in categories:
		if prefix_candidate != None \
			and category.find(prefix_candidate) == 0:
			# prefix_candidate is a prefix of the current category, 
			# so we can skip the current category
			redundant_categories.add(category)
		else:
			prefix_candidate = category
	categories = [category for category in categories if category not in redundant_categories]

	filters = [ops.bw_("category", category) for category in categories]
	if blank:
		filters.append(ops.blank_("category"))
	return ops.or_(*filters)




import logging
import exceptions

class Response(object):
	def __init__(self, body, meta=None):
		self.meta = meta
		self.body = body

		# todo: handle non-"ok" status
		self.status = body.get("status", None)
		self.version = body.get("version", None)
		self.response = body.get("response", {})

		if self.status == "error":
			error = "\"%s\" error: %s" % (body.get("error_type"), body.get("message"))
			raise exceptions.FactualError(error)
	def __repr__(self):
		if len(self.response) > 3:
			response_repr = "%d records in response" % len(self.response)
		else:
			response_repr = self.response.__repr__()
		return "FactualResponse(%s, v%s, %s)" % (self.status, self.version, response_repr)
class ReadResponse(Response):
	pass
class SchemaResponse(Response):
	pass




import logging

import requests
import tables
import factual.common.session

class Session(factual.common.session.BaseSession):
	def read(self, table):
		'''Build a Read request on the provided Table type using this Session. See the requests module for details.'''
		return requests.Read(self, table)
	def input(self, table):
		'''Build an Input request on the provided Table type using this Session. See the requests module for details.'''
		return requests.Input(self, table)
	def rate(self, table):
		'''Build a Rate request on the provided Table type using this Session. See the requests module for details.'''
		return requests.Rate(self, table)
	def duplicates(self, table):
		'''Build a Duplicates request on the provided Table type using this Session. See the requests module for details.'''
		return requests.Duplicates(self, table)
	def schema(self, table):
		'''Build a Schema request on the provided Table type using this Session. See the requests module for details.'''
		return requests.Schema(self, table)




def abstract():
	# http://norvig.com/python-iaq.html
	import inspect
	caller = inspect.getouterframes(inspect.currentframe())[1][3]
	raise NotImplementedError(caller + ' must be implemented in subclass')


class Table(object):
	'''
	Table classes hold information about Factual tables. 

	At the moment, their only purpose is to hold the table's ID for the Session
	to use in building API URLs. 

	To use a new Factual table, subclass Table and set _table_id as a class variable
	holding the new table's ID. 

	'''
	_table_id = None
	def __init__(self, id=None):
		self._id = id
		# todo: query and cache schema for table?
	def __repr__(self):
		return self.id
	def _get_id(self):
		if self._id:
			return self._id
		else:
			return self._table_id
	id = property(_get_id)



'''
Tests for Factual v2 "Server" API wrapper
'''

import nose
from nose.tools import eq_

import factual.v2 as v2
import factual.v3 as v3
import factual.common as common
from factual.v2 import tables

categories = [
	'Sports & Recreation > Gyms & Fitness Centers',
	'Food & Beverage > Bakeries',
	'Arts, Entertainment & Nightlife > Social Clubs',
	'Travel & Tourism > Parking',
	'Food & Beverage > Kosher',
	'Arts, Entertainment & Nightlife > Comedy Clubs',
	'Arts, Entertainment & Nightlife > Karaoke',
	'Arts, Entertainment & Nightlife > Bingo Halls',
	'Food & Beverage > Ethnic Food',
	'Sports & Recreation',
	'Personal Care & Services',
	'Arts, Entertainment & Nightlife > Billiard Parlors & Pool Halls',
	'Travel & Tourism > Resorts',
	'Food & Beverage > Cafes, Coffee Houses & Tea Houses',
	'Arts, Entertainment & Nightlife > Movie Theatres',
	'Food & Beverage > Beer, Wine & Spirits',
	"Food & Beverage > Farmers' Markets",
	'Community & Government > Libraries',
	'Arts, Entertainment & Nightlife > Concert Halls & Theaters',
	'Food & Beverage > Bagels & Donuts',
	'Shopping',
	'Arts, Entertainment & Nightlife > Museums',
	'Arts, Entertainment & Nightlife > Adult Entertainment',
	'Arts, Entertainment & Nightlife > Bars',
	'Arts, Entertainment & Nightlife > Orchestras, Symphonies & Bands',
	'Arts, Entertainment & Nightlife > Bowling Alleys',
	'Food & Beverage > Chocolate',
	'Travel & Tourism > Tourist Attractions',
	'Arts, Entertainment & Nightlife > Night Clubs',
	'Arts, Entertainment & Nightlife > Ticket Sales',
	'Travel & Tourism > Railroads & Trains',
	'Travel & Tourism > Horse-Drawn Vehicles',
	'Sports & Recreation > Zoos, Aquariums & Wildlife Sanctuaries',
	'Community & Government > Day Care & Preschools',
	'Community & Government > Dry Cleaning, Ironing & Laundry',
	'Travel & Tourism > Public Transportation & Transit',
	'Travel & Tourism > Monuments',
	'Food & Beverage > Breweries',
	'Arts, Entertainment & Nightlife > Casinos & Gaming',
	'Food & Beverage > Cheese',
	'Food & Beverage > Restaurants',
	'Food & Beverage > Dessert',
	'Travel & Tourism > Lodging',
	'Legal & Financial > Banking & Financing',
	'Arts, Entertainment & Nightlife > Psychics & Astrologers',
	'Sports & Recreation > Yoga & Pilates',
	'Food & Beverage > Juice Bars & Smoothies',
	'Food & Beverage > Health & Diet Food',
	'Travel & Tourism > Taxi & Car Services',
	'Arts, Entertainment & Nightlife > Hookah Lounges',
	'Travel & Tourism > Wineries & Vineyards',
	'Arts, Entertainment & Nightlife > Jazz & Blues Cafes',
	'Arts, Entertainment & Nightlife',
	'Travel & Tourism > Airports',
	'Travel & Tourism > Historical Sites',
	'Arts, Entertainment & Nightlife > Art Dealers & Galleries',
	'Sports & Recreation > Personal Trainers',
	'Food & Beverage',
	'Community & Government > Religious',
	'Arts, Entertainment & Nightlife > Internet Cafes',
	'Education',
	'Food & Beverage > Ice Cream Parlors',
	'Sports & Recreation > Sports Clubs',
	'Arts, Entertainment & Nightlife > Arcades & Amusement Parks'
	]


API_KEY='deadbeef'

def setup():
	global v2_session,\
	       v3_session,\
	       standard_categories,\
		   standard_categories_with_blank_category,\
		   standard_filters,\
		   standard_filters_with_blank_category
	v2_session = v2.Session(api_key=API_KEY)
	v3_session = v3.Session(api_key=API_KEY)

	standard_categories = common.category_helpers.make_category_filter(categories,blank=False)
	standard_categories_with_blank_category = common.category_helpers.make_category_filter(categories,blank=True)

	standard_filters = standard_categories.copy()
	standard_filters['status'] = 1

	standard_filters_with_blank_category = standard_categories_with_blank_category.copy()
	standard_filters_with_blank_category['status'] = 1

def query_check_helper(query, expected):
	print "query:\n%s\nexpected:\n%s" % (query.get_query(), expected)
	assert query.get_query() == expected 

def check_name_query(session):
	query = session.read(tables.USPOI).filter({"name":"foobar"})
	query_check_helper(query, 'limit=20&filters={%22name%22:%22foobar%22}')

def check_or_query(session):
	query = session.read(tables.USPOI).filter(
											common.shared_filter_helpers.or_(
												common.shared_filter_helpers.bw_("category", "Food"), 
												common.shared_filter_helpers.bw_("category", "Arts")
											)
										).search("Foobar")
	query_check_helper(query, \
		'limit=20&filters={%22$or%22:[{%22category%22:{%22$bw%22:%22Food%22}},{%22category%22:{%22$bw%22:%22Arts%22}}],%22$search%22:%22Foobar%22}')

def check_category_filtering(session):
	query = session.read(tables.USPOI)
	query.filter(standard_categories)
	query_check_helper(query, \
		'''limit=20&filters={%22$or%22:[{%22category%22:{%22$bw%22:%22Arts,+Entertainment+%26+Nightlife%22}},{%22category%22:{%22$bw%22:%22Community+%26+Government+%3E+Day+Care+%26+Preschools%22}},{%22category%22:{%22$bw%22:%22Community+%26+Government+%3E+Dry+Cleaning,+Ironing+%26+Laundry%22}},{%22category%22:{%22$bw%22:%22Community+%26+Government+%3E+Libraries%22}},{%22category%22:{%22$bw%22:%22Community+%26+Government+%3E+Religious%22}},{%22category%22:{%22$bw%22:%22Education%22}},{%22category%22:{%22$bw%22:%22Food+%26+Beverage%22}},{%22category%22:{%22$bw%22:%22Legal+%26+Financial+%3E+Banking+%26+Financing%22}},{%22category%22:{%22$bw%22:%22Personal+Care+%26+Services%22}},{%22category%22:{%22$bw%22:%22Shopping%22}},{%22category%22:{%22$bw%22:%22Sports+%26+Recreation%22}},{%22category%22:{%22$bw%22:%22Travel+%26+Tourism+%3E+Airports%22}},{%22category%22:{%22$bw%22:%22Travel+%26+Tourism+%3E+Historical+Sites%22}},{%22category%22:{%22$bw%22:%22Travel+%26+Tourism+%3E+Horse-Drawn+Vehicles%22}},{%22category%22:{%22$bw%22:%22Travel+%26+Tourism+%3E+Lodging%22}},{%22category%22:{%22$bw%22:%22Travel+%26+Tourism+%3E+Monuments%22}},{%22category%22:{%22$bw%22:%22Travel+%26+Tourism+%3E+Parking%22}},{%22category%22:{%22$bw%22:%22Travel+%26+Tourism+%3E+Public+Transportation+%26+Transit%22}},{%22category%22:{%22$bw%22:%22Travel+%26+Tourism+%3E+Railroads+%26+Trains%22}},{%22category%22:{%22$bw%22:%22Travel+%26+Tourism+%3E+Resorts%22}},{%22category%22:{%22$bw%22:%22Travel+%26+Tourism+%3E+Taxi+%26+Car+Services%22}},{%22category%22:{%22$bw%22:%22Travel+%26+Tourism+%3E+Tourist+Attractions%22}},{%22category%22:{%22$bw%22:%22Travel+%26+Tourism+%3E+Wineries+%26+Vineyards%22}}]}''')

def check_uuid_lookup(session):
	query = session.read(tables.USPOI)
	query.filter({'factual_id': 'b3b08e30-8e3b-012e-06c9-003048c87378'})
	query_check_helper(query, 'limit=20&filters={%22factual_id%22:%22b3b08e30-8e3b-012e-06c9-003048c87378%22}')

#### Shared test generators
#
# Run these tests/checks against both the v2 and v3 apis.
#

# Should have identical results.
def test_name_query():
	for session in [ v2_session ]:
		yield check_name_query, session

def test_or_query():
	for session in [ v2_session ]:
		yield check_or_query, session

def test_category_filtering():
	for session in [ v2_session ]:
		yield check_category_filtering, session

def test_uuid_lookup():
	for session in [ v2_session ]:
		yield check_uuid_lookup, session

# These will differ as noted:
def test_search_query():
	v2_query = v2_session.read(tables.USPOI).search("coffee and tea")
	v2_expected = 'limit=20&filters={%22$search%22:%22coffee+and+tea%22}'
	yield query_check_helper, v2_query, v2_expected

#	v3_query = v3_session.read(tables.USPOI).search("coffee and tea")
#	v3_expected = 'limit=20&q=coffee+and+tea'
#	yield query_check_helper, v3_query, v3_expected

def test_url_from_table_type():
	v2_query = v2_session.read(tables.USPOI)
	eq_(v2_session.get_url(v2_query), 'http://api.factual.com/v2/tables/s4OOB4/read?api_key=deadbeef&limit=20')
def test_url_from_table_instance():
	v2_query = v2_session.read(tables.USPOI())
	eq_(v2_session.get_url(v2_query), 'http://api.factual.com/v2/tables/s4OOB4/read?api_key=deadbeef&limit=20')
def test_url_from_table_parent_instance():
	v2_query = v2_session.read(common.util.Table("s4OOB4"))
	eq_(v2_session.get_url(v2_query), 'http://api.factual.com/v2/tables/s4OOB4/read?api_key=deadbeef&limit=20')
def test_url_from_table_string():
	v2_query = v2_session.read("s4OOB4")
	eq_(v2_session.get_url(v2_query), 'http://api.factual.com/v2/tables/s4OOB4/read?api_key=deadbeef&limit=20')

def test_v3_url_from_table_parent_instance():
	v3_query = v3_session.read(common.util.Table("places"))
	eq_(v3_session.get_url(v3_query), 'http://api.v3.factual.com/t/places/read?limit=20&KEY=deadbeef')
def test_v3_url_from_table_string():
	v3_query = v3_session.read("places")
	eq_(v3_session.get_url(v3_query), 'http://api.v3.factual.com/t/places/read?limit=20&KEY=deadbeef')

#### v2 only

def test_v2_geo_query():
	query = v2_session.read(tables.USPOI).search("coffee").within(40.7353,-73.9912,1000)
	query_check_helper(query, \
		'limit=20&filters={%22$loc%22:{%22$within%22:{%22$center%22:[[40.735300,-73.991200],1000]}},%22$search%22:%22coffee%22}')

	print v2_session.get_url(query)

def test_v2_category_and_radius_query():
	query = v2_session.read(tables.USPOI)
	query.filter(standard_filters_with_blank_category)
	query.filter(v2.filter_helpers.within_(40.7353, -73.9912, 1500000))
	query_check_helper(query, \
		'''limit=20&filters={%22$or%22:[{%22category%22:{%22$bw%22:%22Arts,+Entertainment+%26+Nightlife%22}},{%22category%22:{%22$bw%22:%22Community+%26+Government+%3E+Day+Care+%26+Preschools%22}},{%22category%22:{%22$bw%22:%22Community+%26+Government+%3E+Dry+Cleaning,+Ironing+%26+Laundry%22}},{%22category%22:{%22$bw%22:%22Community+%26+Government+%3E+Libraries%22}},{%22category%22:{%22$bw%22:%22Community+%26+Government+%3E+Religious%22}},{%22category%22:{%22$bw%22:%22Education%22}},{%22category%22:{%22$bw%22:%22Food+%26+Beverage%22}},{%22category%22:{%22$bw%22:%22Legal+%26+Financial+%3E+Banking+%26+Financing%22}},{%22category%22:{%22$bw%22:%22Personal+Care+%26+Services%22}},{%22category%22:{%22$bw%22:%22Shopping%22}},{%22category%22:{%22$bw%22:%22Sports+%26+Recreation%22}},{%22category%22:{%22$bw%22:%22Travel+%26+Tourism+%3E+Airports%22}},{%22category%22:{%22$bw%22:%22Travel+%26+Tourism+%3E+Historical+Sites%22}},{%22category%22:{%22$bw%22:%22Travel+%26+Tourism+%3E+Horse-Drawn+Vehicles%22}},{%22category%22:{%22$bw%22:%22Travel+%26+Tourism+%3E+Lodging%22}},{%22category%22:{%22$bw%22:%22Travel+%26+Tourism+%3E+Monuments%22}},{%22category%22:{%22$bw%22:%22Travel+%26+Tourism+%3E+Parking%22}},{%22category%22:{%22$bw%22:%22Travel+%26+Tourism+%3E+Public+Transportation+%26+Transit%22}},{%22category%22:{%22$bw%22:%22Travel+%26+Tourism+%3E+Railroads+%26+Trains%22}},{%22category%22:{%22$bw%22:%22Travel+%26+Tourism+%3E+Resorts%22}},{%22category%22:{%22$bw%22:%22Travel+%26+Tourism+%3E+Taxi+%26+Car+Services%22}},{%22category%22:{%22$bw%22:%22Travel+%26+Tourism+%3E+Tourist+Attractions%22}},{%22category%22:{%22$bw%22:%22Travel+%26+Tourism+%3E+Wineries+%26+Vineyards%22}},{%22category%22:{%22$blank%22:true}}],%22status%22:1,%22$loc%22:{%22$within%22:{%22$center%22:[[40.735300,-73.991200],1500000]}}}''')


#### v3 only

def test_v3_geo_query_compat():
	# old style "within" compat
	query = v3_session.read("places").search("coffee").within(40.7353,-73.9912,1000)
	query_check_helper(query, 'q=coffee&geo={%22$circle%22:{%22$center%22:[40.735300,-73.991200],%22$meters%22:1000}}&limit=20')

	print v3_session.get_url(query)

#def test_v3_geo_query_new_style():
#	# new style circle
#	query = v3_session.read("places").search("coffee").filter({"$circle":{"$center":[40.7353,-73.9912], "$meters":1000}})



import logging

from factual.common import exceptions
from factual.common.responses import Response

class V2ReadResponse(Response):
	def __init__(self, body, meta=None):
		Response.__init__(self, body, meta)
		self.data = self.response.get("data", None)
		self.fields = self.response.get("fields", None)
		self._records = None
	def _process_row(self, row):
		# todo: wrap record in custom dict subclass? keep field references?
		record = {}
		for i in range(len(self.fields)):
			record[self.fields[i]] = row[i]
		return record
	def _get_records(self):
		return [self._process_row(row) for row in self.data]
	def records(self):
		'''Return a list of dicts corresponding to the data returned by Factual.'''
		if self._records == None:
			self._records = self._get_records()
		return self._records
class V2SchemaResponse(Response):
	def __init__(self, body, meta=None):
		Response.__init__(self, body, meta)
		self.schema = self.body.get("schema", {})




'''
Factual Server API  Wrapper
===========================

This package wraps the Factual "Server" API. Its query style is SQLAlchemy-inspired and designed
to make it easy to build "read" requests by chaining filter calls together. 

API actions other than "read" are supported via the same syntax for consistency. 

Basic Usage:

    from factual import *
	from factual.v2.tables import USPOI,USLocalPlaypen
    s = Session(api_key="deadbeef")
    my_place = s.read(USPOI).search("coffee").run().records()[0]
    
    # you can also build requests one piece at a time:
    query = s.read(USPOI)
    query.filter({"name": "Foobar"})
	if test_address != None:
		query.filter({"address": test_address})
    response = query.run()
    records = response.records()
    
    # limiting categories and using the filter helper functions:
    from factual.v2.filter_helpers import *
    q = s.read(USPOI).filter(or_(bw_("category", "Food"), bw_("category", "Arts")))
    q.search("Foobar")
    records = q.run().records()
    
    # geo queries:
    coffee_places = s.read(USPOI).search("coffee").within(40.7353,-73.9912,1000).run().records()
    
    # modify a record in the Playpen:
    p = s.read(USLocalPlaypen).search("coffee").count(1).run().records()[0]
    p['address'] += "/Foobar"
    s.input(USLocalPlaypen).values(p).comment("Silly update test").run()

See also the session.Session and requests.Read documentation.
'''

import logging
logging.basicConfig(level=logging.DEBUG)

from session import Session
import filter_helpers
import factual.common.category_helpers as category_helpers
import tables





import factual.common.requests
from factual.common.requests import Read
import responses
import filter_helpers

class Read(factual.common.requests.Read):
	'''Override response_type with correct ReadResponse type'''
	response_type = responses.V3ReadResponse
	def __init__(self, session, table):
		factual.common.requests.Read.__init__(self, session, table)
		self._geo = {}
		self._search = None
	def get_params(self):
		params = factual.common.requests.Read.get_params(self)
		params["geo"] = self.unparse_dict(self._geo) if self._geo != {} else None
		params["q"] = self._search
		return params
	def search(self, search_term):
		'''
		Apply a "q=..." search filter. Separate terms with commas for OR queries. 
		Subsequent search() calls replace any previous search terms on this query.
		'''
		self._search = search_term
		return self
	def within(self, lat, lon, radius):
		'''
		Apply a geo=$circle/$center filter. Radius is in meters. Subsequent within() 
		calls replace earlier geo filters on this query.
		'''
		self._geo = { "$circle": { 
						"$center": [
							factual.common.shared_filter_helpers.GeoScalar(lat), 
							factual.common.shared_filter_helpers.GeoScalar(lon), 
							], 
						"$meters": radius 
						}
					}
		return self




import datetime
import time

class GeoScalar(float):
	'''Class to wrap floats used for geo coords in order to get pretty formatting from simplejson.'''
	def __repr__(self):
		return "%0.6f" % (self)

### Factual filter functions
def _prep_factual_term(term):
	# ensure dates are converted to epoch
	if type(term)==datetime.datetime:
		term = time.mktime(term)
	return term

# generic filters
def or_(*args):
	return {"$or": args}
def and_(*args):
	return {"$and": args}
def field_filter(op, field, term):
	return {field: {op: _prep_factual_term(term)}}
# text and numeric ops
def eq_(field, term):
	return {field: _prep_factual_term(term)}
def neq_(field, term):
	return field_filter("$neq", field, term)
def in_(field, terms):
	if type(terms) != list:
		terms = [terms]
	return field_filter("$in", field, terms)
def blank_(field):
	return field_filter("$blank", field, True)
def not_blank_(field):
	return field_filter("$blank", field, False)
# text ops
def bw_(field, term):
	return field_filter("$bw", field, term)
def search_(term):
	return {"$search": term}

# numeric ops
def gt_(field, term):
	return field_filter("$gt", field, term)
def gte_(field, term):
	return field_filter("$gte", field, term)
def lt_(field, term):
	return field_filter("$lt", field, term)
def lte_(field, term):
	return field_filter("$lte", field, term)




import factual.common.requests
from factual.common.requests import Read
import responses
import filter_helpers

class Read(factual.common.requests.Read):
	'''Override response_type with correct ReadResponse type'''
	response_type = responses.V2ReadResponse
	def search(self, search_term):
		'''Convenience method to apply a {"$search":...} filter'''
		return self.filter(filter_helpers.search_(search_term))
	def within(self, lat, lon, radius):
		'''Convenience method to apply a $loc/$within/$center filter. Radius is in meters.'''
		return self.filter(filter_helpers.within_(lat, lon, radius))

class Input(factual.common.requests.WriteRequest):
	action = "input"
	def __init__(self, session, table):
		Request.__init__(self, session, table)
		self._values = None
		self._comments = None
		self._source = None
		self._subject_key = None
		self._token = None
	def get_query(self):
		# todo: decide whether subject key removal here makes sense
		if 'subject_key' in self._values:
			subject_key = self._values['subject_key']
			del self._values['subject_key']
			if self._subject_key == None:
				self._subject_key = subject_key
		params = {
			"values": self.unparse_dict(self._values),
			"comments": self._comments,
			"source": self._source,
			"subject_key": self._subject_key,
			"token": self._token
			}
		return self.join_params(params)
	def values(self, values):
		'''
		Set the values dictionary to write to the table.

		N.B. Because the subject_key is included in row dicts returned by this library
		     but is also not allowed as a field in a values sumission to Factual's input 
			 action, if 'subject_key' is in the submitted values dict, it will be removed
			 and passed as the subject_key parameter to this edit, unless a different, 
			 non-None subject_key is explicitly specified. 

		     This will make this edit an update rather than an insert. 
		'''
		self._values = values.copy()
		return self
	def source(self, source):
		'''Set a source citation for this edit.'''
		self._source = source
		return self
	def comments(self, comments):
		'''Set a comment for this edit.'''
		self._comments = comments
		return self
	def subject_key(self, subject_key):
		'''
		Set a subject_key of the row to be update.  Setting a subject key causes the 
		input action to be treated as an update rather than an insert.
		'''
		self._subject_key = subject_key
		return self
	def token(self, token):
		'''Set a user's id token as the author of this edit.'''
		self._token = token
		return self
class Rate(factual.common.requests.WriteRequest):
	action = "rate"
	def __init__(self, session, table):
		Request.__init__(self, session, table)
		self._rating = None
		self._subject_key = None
	def get_query(self):
		return self.join_params({"rating": self._rating, "subject_key": self._subject_key})
	def rating(self, rating):
		'''Set the rating for the specified row.'''
		self._rating = rating
		return self
	def subject_key(self, subject_key):
		'''Set the subject_key of the row to be rated.'''
		self._subject_key = subject_key
		return self
class Duplicates(factual.common.requests.Request):
	action = "duplicates"
	# todo: create DuplicateResponse 
	#response_type = responses.V2DuplicateResponse
	def __init__(self, session, table):
		Request.__init__(self, session, table)
		self._values = None
	def get_query(self):
		return self.join_params({"values": self.unparse_dict(self._values)})
	def values(self, values):
		'''Set the values dictionary to be compared against the Factual table.'''
		self._values = values.copy()
		return self




import logging
import urllib
import unidecode
try:
	import simplejson as json
except ImportError:
	import json

import responses
import shared_filter_helpers
import util

class Request(object):
	response_type = responses.Response
	def __init__(self, session, table):
		self.session = session
		# instantiate the table class if we're passed a type rather than an instance (or a string)
		self.table = table() if isinstance(table, type) else table
	def run(self, async=False):
		'''
		Perform the Factual API HTTP request. 

		By default, process and return the API response wrapped in a
		FactualResponse object. 

		Pass `async = True` to perform the request asynchronously; in 
		async mode, `run` returns a function, get_response, that will 
		process and return the response when called. 
		
		Additionally, if the asynchttp module is available and async is
		True, the initial HTTP request will not block, and get_response
		will be returned immediately.  get_response function will block 
		when called if the HTTP requst is not yet complete. 
		'''
		return self.session.run(self, async)
	def make_response(self, *args, **kwargs):
		response = self.response_type(*args, **kwargs)
		return response
	def unparse_dict(self, d):
		return unidecode.unidecode(\
			json.dumps(d,ensure_ascii=False, \
					separators=(',',':')))
	def join_params(self, params, suppress_null=True):
		''' 
		Custom alternative to urllib.urlencode that keeps URL length
		down by not escaping certain special characters that Factual
		can tolerate unescaped. This is especially useful for the 
		JSON special chars {}[],:. 

		The full list of usually-escaped-but-ok-for-factual characters is:

		    / { } ' $ : , [ ]

		The API used to tolerate doublequotes, but this seems to have 
		changed. 

		suppress_null removes parametrs which are set to None.
		'''
		# ensure all params and values are properly urlencoded
		# since we may have numbers, etc. here, cast to unicode and encode back to utf-8
		return "&".join([
					("%s=%s" \
						% (
							urllib.quote_plus(unicode(k).encode('utf-8'), '/{}\'$:,[]'),
							urllib.quote_plus(unicode(v).encode('utf-8'), '/{}\'$:,[]')
						)
					) 
					for k,v 
					in params.iteritems()
					if not suppress_null or v != None
				])

				
	def get_query(self): util.abstract()
class WriteRequest(Request):
	pass
class Read(Request):
	'''
	Query object used to build Factual read requests. 

	Read objects should be created by calling read(<Table type>) on a Session object:
		s = Session()
		query = s.read("places")
	
	Creating a query object this way automatically binds it to the Session and Table.
	
	A Read query can be customized with filters and other paramters. Read objects'
	functions return the same Read object, making them chainable:
		query = s.read("places")
		query = query.filter({"name":"foobar"}).filter({"address":"123 Main St."})
	
	Once a query is ready, call run to get a Result object. Read queries return
	ReadResult objects which process the returned data rows into dicts:
		res = query.run()
		records = res.records()

	or, chained:
		records = s.read(tables.USPOI).filter({"name":"foobar"}).run().records()
	'''
	action = "read"
	response_type = responses.ReadResponse
	def __init__(self, session, table):
		Request.__init__(self, session, table)
		self._filters = {}
		self._count = 20
		self._page = 1
	def get_query(self):
		return self.join_params(self.get_params())
	def get_params(self):
		filters = dict([(k,v) for k,v in self._filters.iteritems() if v != None])

		params = {
			"filters": self.unparse_dict(filters) if filters != {} else None,
			"limit": self._count
		}

		offset = self._count * (self._page - 1)
		if offset > 0:
			params["offset"] = offset

		return params
	def count(self, count):
		'''Set the number of results to return per page. N.B. Factual caps this at 50.'''
		self._count = count 
		return self #chainable
	def page(self, page):
		'''
		Set the page to return. Factual caps the number of rows requestable, or page * count, 
		at 500.
		'''
		self._page = page
		return self
	
	def filter(self, filter):
		'''
		Add a filter to this Read request.

		The filter should be a dict containing a valid Factual filter tree.
		filter() can be called multiple times. Filter dict combination is handled
		with dict.update(), so subsequent conflicting keys will overwrite previous 
		values. 

		The helper functions in the filter_helpers module can be used to generate
		valid Facutal filter dicts.

		To remove a filter already applied, pass a dict containing the key to remove 
		with None as its value.

		Examples:
			- s.read(tables.USPOI).filter({"name":{"$bw":"foo"}})
			- Use filter_helpers:
				from filter_helpers import *
				s.read(tables.USPOI).filter(bw_("name", "foo"))
			- Set name and $search filters:
				s.read(tables.USPOI).filter({"name":"foobar"}).filter({"$search":"baz"})
			- Overwrite previous value; name:baz prevails
				s.read(tables.USPOI).filter({"name":"foobar"}).filter({"name":"baz"}) 
			- Remove previous filter:
				s.read(tables.USPOI).filter({"name":"foobar"}).filter({"name":None})

		'''
		self._filters.update(filter)
		return self # chainable
	def name(self, term):
		'''Convenience method to apply a {"name":...} filter'''
		return self.filter(shared_filter_helpers.eq_("name", term))
class Schema(Request):
	action = "schema"
	response_type = responses.SchemaResponse
	def get_query(self):
		# schema action takes no parameters
		return ""




#!/usr/bin/env python

from setuptools import setup, find_packages
import sys
import subprocess

required_modules = [
	"simplejson",
	"httplib2",
	"unidecode",
	"oauth2"
	]

readme = ""
try:
	with open("README.md", "rb") as f:
		readme = f.read()

	# avoid ReST at all costs, but make sure we have it for PyPi
	proc = subprocess.Popen(
		["pandoc", "-fmarkdown", "-trst", "README.md"], 
		stdout=subprocess.PIPE
	)
	readme = proc.stdout.read()
except:
	if len(sys.argv) >=2:
		for command in ["register", "upload"]:
			if command in sys.argv[1:]:
				raise Exception("Unable to convert Markdown README to ReST for upload to PyPi. Do you have pandoc installed?")

setup(
	name="factual",
	version="0.1.3",
	description="Wrapper for the Factual HTTP API.",
	author="Christopher H. Casebeer",
	author_email="",
	url="https://github.com/casebeer/factual",

	packages=find_packages(exclude='tests'),
	install_requires=required_modules,

	tests_require=["nose"],
	test_suite="nose.collector",

	long_description=readme,
	classifiers=[
		"License :: OSI Approved :: BSD License",
		"Intended Audience :: Developers",
	]
)




import logging

from factual.common import exceptions
from factual.common.responses import Response

class V3ReadResponse(Response):
	def __init__(self, body, meta=None):
		Response.__init__(self, body, meta)
		self._data = self.response.get("data", None)
	def records(self):
		'''Return a list of dicts corresponding to the data returned by Factual.'''
		return self._data
class V3SchemaResponse(Response):
	def __init__(self, body, meta=None):
		Response.__init__(self, body, meta)
		self.view = self.response.get("view", {})




'''
v2 Tables
'''

from factual.common.util import Table

class USPOI(Table):
	'''
	US POI and Business Listings table
	http://www.factual.com/t/s4OOB4/US_POI_and_Business_Listings
	'''
	_table_id = "s4OOB4"

class USLocalPlaypen(Table):
	'''
	Playpen for U.S. local data table for experimenting with API calls
	http://www.factual.com/t/Nj0JN3/POI_and_Local_sandbox_Los_Angeles_CA
	'''
	_table_id = "Nj0JN3"

class AULocalPlaypen(Table):
	'''
	Playpen for Australian local data
	http://www.factual.com/t/eN7l2X/Sydney_POI_and_Business_Listings_API_PLAYGROUND
	'''
	_table_id = "eN7l2X"



import logging
try:
	import asynchttp as http
except ImportError:
	import httplib2 as http

try:
	import simplejson as json
except ImportError:
	import json

class BaseSession(object):
	'''
	Main entry point to access the Factual "Server" API. 

	Session instances hold API level configuration; specifically, the api_key and api_base URL to use.

	Session instances also create read, input, rate, duplicates, or schema request objects that
	can be used to build and run requests against the Facutal API. See the requests module for 
	details on customizing API requests. 

	Session instances should be the *only* way you create request objects. 

	Examples:

		s = Session(api_key="deadbeef")
		records = s.read("places").search("coffee").run().records()
	
	Note that "Session" is meant loosely in the "login session" sense, *not* the "transaction" sense.
	'''
	_url_pat = "%(api_base)s/tables/%(table_id)s/%(action)s?api_key=%(api_key)s&%(query)s"
	def __init__(self, api_key=None, api_base="http://api.factual.com/v2"):
		'''Provide an api_key to use with the Factual API.'''
		self.api_key = api_key
		self.api_base = api_base
	def get_url(self, request):
		return self._url_pat % {
				'api_base': self.api_base,
				'table_id': request.table,
				'action':   request.action,
				'api_key':  self.api_key,
				'query':    request.get_query()
			}
	def get_headers(self, request, defaults={}):
		return {}.update(defaults)
	def run(self, request, async=False):
		url = self.get_url(request)
		headers = self.get_headers(request)

		logging.debug(url)
		logging.debug(headers)

		h = http.Http()
		http_response, http_body = h.request(url, headers=headers)

		def get_response():
			'''
			Process and return the HTTP response from Factual.

			Performs the post-request processing needed to handle a Factual 
			response. Broken into a separate function so the post processing
			can be deferred, e.g. for use with asynchttp in place of
			httplib2. 

			When the asynchttp module is installed, this call will block
			if the HTTP request to Factual is not yet complete. 
			'''
			# todo: timing and other metrics
			meta = {}
			response = request.make_response(json.loads(str(http_body)), meta=meta)
			return response

		if async:
			return get_response
		else:
			return get_response()

